/*
Copyright 2021 The Kubernetes Authors All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package systemstatsmonitor

import (
	"fmt"

	"github.com/golang/glog"
	"github.com/prometheus/procfs"
	ssmtypes "k8s.io/node-problem-detector/pkg/systemstatsmonitor/types"
	"k8s.io/node-problem-detector/pkg/util/metrics"
)

type processCollector struct {
	mRSSUsage            *metrics.Int64Metric
	mVMUsage             *metrics.Int64Metric
	mReadBytesCount      *metrics.Int64Metric
	mWriteBytesCount     *metrics.Int64Metric
	mReadOperationCount  *metrics.Int64Metric
	mWriteOperationCount *metrics.Int64Metric
	mCpuPercentage       *metrics.Float64Metric

	config *ssmtypes.ProcessStatsConfig
}

func NewProcessCollectorOrDie(processStatsConfig *ssmtypes.ProcessStatsConfig) *processCollector {
	pc := processCollector{config: processStatsConfig}
	var err error

	if processStatsConfig.MetricsConfigs[string(metrics.ProcessRSSUsage)].DisplayName != "" {
		pc.mRSSUsage, err = metrics.NewInt64Metric(
			metrics.ProcessRSSUsage,
			processStatsConfig.MetricsConfigs[string(metrics.ProcessRSSUsage)].DisplayName,
			"Resident memory usage of the given process.",
			"1",
			metrics.LastValue,
			[]string{pidLabel, processLabel, ownerLabel})
		if err != nil {
			glog.Fatalf("Error initializing metric for process/rss_usage: %v", err)
		}
	}

	if processStatsConfig.MetricsConfigs[string(metrics.ProcessVMSUsage)].DisplayName != "" {
		pc.mVMUsage, err = metrics.NewInt64Metric(
			metrics.ProcessVMSUsage,
			processStatsConfig.MetricsConfigs[string(metrics.ProcessVMSUsage)].DisplayName,
			"Resident memory usage of the given process.",
			"1",
			metrics.LastValue,
			[]string{pidLabel, processLabel, ownerLabel})
		if err != nil {
			glog.Fatalf("Error initializing metric for process/vm_usage: %v", err)
		}
	}

	if processStatsConfig.MetricsConfigs[string(metrics.ProcessCPUPercent)].DisplayName != "" {
		pc.mCpuPercentage, err = metrics.NewFloat64Metric(
			metrics.ProcessCPUPercent,
			processStatsConfig.MetricsConfigs[string(metrics.ProcessCPUPercent)].DisplayName,
			"CPU Percentage of a given process.",
			"1",
			metrics.LastValue,
			[]string{pidLabel, processLabel, ownerLabel})
		if err != nil {
			glog.Fatalf("Error initializing metric for process/vm_usage: %v", err)
		}
	}

	if processStatsConfig.MetricsConfigs[string(metrics.ProcessReadCount)].DisplayName != "" {
		pc.mReadOperationCount, err = metrics.NewInt64Metric(
			metrics.ProcessReadCount,
			processStatsConfig.MetricsConfigs[string(metrics.ProcessReadCount)].DisplayName,
			"Number of read operations.",
			"1",
			metrics.Sum,
			[]string{pidLabel, processLabel, ownerLabel})
		if err != nil {
			glog.Fatalf("Error initializing metric for process/vm_usage: %v", err)
		}
	}

	if processStatsConfig.MetricsConfigs[string(metrics.ProcessWriteCount)].DisplayName != "" {
		pc.mWriteOperationCount, err = metrics.NewInt64Metric(
			metrics.ProcessWriteCount,
			processStatsConfig.MetricsConfigs[string(metrics.ProcessWriteCount)].DisplayName,
			"Number of write operations.",
			"1",
			metrics.Sum,
			[]string{pidLabel, processLabel, ownerLabel})
		if err != nil {
			glog.Fatalf("Error initializing metric for process/vm_usage: %v", err)
		}
	}

	if processStatsConfig.MetricsConfigs[string(metrics.ProcessReadBytes)].DisplayName != "" {
		pc.mReadBytesCount, err = metrics.NewInt64Metric(
			metrics.ProcessReadBytes,
			processStatsConfig.MetricsConfigs[string(metrics.ProcessReadBytes)].DisplayName,
			"Number of read bytes.",
			"1",
			metrics.Sum,
			[]string{pidLabel, processLabel, ownerLabel})
		if err != nil {
			glog.Fatalf("Error initializing metric for process/vm_usage: %v", err)
		}
	}

	if processStatsConfig.MetricsConfigs[string(metrics.ProcessWriteBytes)].DisplayName != "" {
		pc.mWriteBytesCount, err = metrics.NewInt64Metric(
			metrics.ProcessWriteBytes,
			processStatsConfig.MetricsConfigs[string(metrics.ProcessWriteBytes)].DisplayName,
			"Number of write bytes.",
			"1",
			metrics.Sum,
			[]string{pidLabel, processLabel, ownerLabel})
		if err != nil {
			glog.Fatalf("Error initializing metric for process/vm_usage: %v", err)
		}
	}

	return &pc
}

func (pc *processCollector) collect() {
	processesInfo, err := procfs.AllProcs()
	if err != nil {
		glog.Errorf("Failed to retrieve processes of the host: %v", err)
		return
	}

	if pc.mCpuPercentage == nil {
		return
	}
	if pc.mWriteBytesCount == nil {
		return
	}
	if pc.mReadBytesCount == nil {
		return
	}

	if pc.mWriteOperationCount == nil {
		return
	}
	if pc.mReadOperationCount == nil {
		return
	}
	if pc.mRSSUsage == nil {
		return
	}
	if pc.mVMUsage == nil {
		return
	}

	for _, process := range processesInfo {
		pid := process.PID
		pid_process, err := procfs.NewProc(pid)
		pid_process_stat, _ := pid_process.Stat()
		processName, _ := pid_process.Comm()

		if err != nil {
			glog.Infof("Failed to retrieve the process information: %v", err)
			return
		}


		pc.mCpuPercentage.Record(map[string]string{pidLabel: fmt.Sprintf("%v", pid), processLabel: processName}, pid_process_stat.CPUTime())
		glog.Infof("CPU pid %d: , processName: %s, metricValue: %f", pid, processName, pid_process_stat.CPUTime())

		pc.mRSSUsage.Record(map[string]string{pidLabel: fmt.Sprintf("%v", pid), processLabel: processName}, int64(pid_process_stat.ResidentMemory()))
		glog.Infof("RSS pid %d: , processName: %s, metricValue: %d", pid, processName, pid_process_stat.ResidentMemory())

		pc.mVMUsage.Record(map[string]string{pidLabel: fmt.Sprintf("%v", pid), processLabel: processName}, int64(pid_process_stat.VirtualMemory()))
		glog.Infof("VMM pid %d: , processName: %s, metricValue: %d", pid, processName, pid_process_stat.VirtualMemory())

	}

}
